@@ -27,8 +27,27 @@
 #include <pkcs11_int.h>
 #include <gnutls_sig.h>
 #include <gnutls_pk.h>
+#include <fips.h>
 #include <p11-kit/uri.h>
 
+/* In case of a fork, it will invalidate the open session
+ * in privkey */
+#define PKCS11_CHECK_INIT_PRIVKEY(k) \
+	ret = _gnutls_pkcs11_check_init(); \
+	if (ret < 0) \
+		return gnutls_assert_val(ret); \
+	if (ret == 1) \
+		memset(&k->sinfo, 0, sizeof(k->sinfo))
+
+#define CHECK_SHANDLE(rv, key, sinfo) \
+		if (rv == CKR_SESSION_HANDLE_INVALID && sinfo == &key->sinfo) { \
+			if (key->sinfo.init != 0) { \
+				pkcs11_close_session(&key->sinfo); \
+				key->sinfo.init = 0; \
+				goto restart; \
+			} \
+		}
+
 struct gnutls_pkcs11_privkey_st {
 	gnutls_pk_algorithm_t pk_algorithm;
 	unsigned int flags;
@@ -51,6 +70,8 @@ struct gnutls_pkcs11_privkey_st {
  **/
 int gnutls_pkcs11_privkey_init(gnutls_pkcs11_privkey_t * key)
 {
+        FAIL_IF_LIB_ERROR;
+
 	*key = gnutls_calloc(1, sizeof(struct gnutls_pkcs11_privkey_st));
 	if (*key == NULL) {
 		gnutls_assert();
@@ -215,6 +236,9 @@ _gnutls_pkcs11_privkey_sign_hash(gnutls_pkcs11_privkey_t key,
 	struct pkcs11_session_info *sinfo;
 	ck_object_handle_t obj;
 
+	PKCS11_CHECK_INIT_PRIVKEY(key);
+
+ restart:
 	if (key->sinfo.init != 0) {
 		sinfo = &key->sinfo;
 		obj = key->obj;
@@ -233,6 +257,7 @@ _gnutls_pkcs11_privkey_sign_hash(gnutls_pkcs11_privkey_t key,
 	rv = pkcs11_sign_init(sinfo->module, sinfo->pks, &mech, obj);
 	if (rv != CKR_OK) {
 		gnutls_assert();
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = pkcs11_rv_to_err(rv);
 		goto cleanup;
 	}
@@ -242,6 +267,7 @@ _gnutls_pkcs11_privkey_sign_hash(gnutls_pkcs11_privkey_t key,
 			 NULL, &siglen);
 	if (rv != CKR_OK) {
 		gnutls_assert();
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = pkcs11_rv_to_err(rv);
 		goto cleanup;
 	}
@@ -253,6 +279,7 @@ _gnutls_pkcs11_privkey_sign_hash(gnutls_pkcs11_privkey_t key,
 			 tmp.data, &siglen);
 	if (rv != CKR_OK) {
 		gnutls_assert();
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = pkcs11_rv_to_err(rv);
 		goto cleanup;
 	}
@@ -319,7 +346,10 @@ int gnutls_pkcs11_privkey_status(gnutls_pkcs11_privkey_t key)
 	struct pkcs11_session_info *sinfo;
 	ck_object_handle_t obj;
 	struct ck_session_info session_info;
+	
+	PKCS11_CHECK_INIT_PRIVKEY(key);
 
+ restart:
 	if (key->sinfo.init != 0) {
 		sinfo = &key->sinfo;
 		obj = key->obj;
@@ -331,6 +361,7 @@ int gnutls_pkcs11_privkey_status(gnutls_pkcs11_privkey_t key)
 
 	rv = (sinfo->module)->C_GetSessionInfo(sinfo->pks, &session_info);
 	if (rv != CKR_OK) {
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = 0;
 		goto cleanup;
 	}
@@ -368,6 +399,8 @@ gnutls_pkcs11_privkey_import_url(gnutls_pkcs11_privkey_t pkey,
 	ck_key_type_t key_type;
 	struct pkcs11_session_info sinfo;
 
+	PKCS11_CHECK_INIT;
+
 	memset(&sinfo, 0, sizeof(sinfo));
 
 	ret = pkcs11_url_to_info(url, &pkey->info);
@@ -458,6 +491,9 @@ _gnutls_pkcs11_privkey_decrypt_data(gnutls_pkcs11_privkey_t key,
 	struct pkcs11_session_info _sinfo;
 	struct pkcs11_session_info *sinfo;
 
+	PKCS11_CHECK_INIT_PRIVKEY(key);
+
+ restart:
 	if (key->sinfo.init != 0) {
 		sinfo = &key->sinfo;
 		obj = key->obj;
@@ -479,6 +515,7 @@ _gnutls_pkcs11_privkey_decrypt_data(gnutls_pkcs11_privkey_t key,
 	rv = pkcs11_decrypt_init(sinfo->module, sinfo->pks, &mech, obj);
 	if (rv != CKR_OK) {
 		gnutls_assert();
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = pkcs11_rv_to_err(rv);
 		goto cleanup;
 	}
@@ -488,6 +525,7 @@ _gnutls_pkcs11_privkey_decrypt_data(gnutls_pkcs11_privkey_t key,
 			    ciphertext->size, NULL, &siglen);
 	if (rv != CKR_OK) {
 		gnutls_assert();
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = pkcs11_rv_to_err(rv);
 		goto cleanup;
 	}
@@ -500,6 +538,7 @@ _gnutls_pkcs11_privkey_decrypt_data(gnutls_pkcs11_privkey_t key,
 	if (rv != CKR_OK) {
 		gnutls_free(plaintext->data);
 		gnutls_assert();
+		CHECK_SHANDLE(rv, key, sinfo);
 		ret = pkcs11_rv_to_err(rv);
 		goto cleanup;
 	}
@@ -509,7 +548,7 @@ _gnutls_pkcs11_privkey_decrypt_data(gnutls_pkcs11_privkey_t key,
 	ret = 0;
 
       cleanup:
-	if (key->sinfo.init == 0)
+	if (sinfo != &key->sinfo)
 		pkcs11_close_session(sinfo);
 
 	return ret;
@@ -617,6 +656,8 @@ gnutls_pkcs11_privkey_generate2(const char *url, gnutls_pk_algorithm_t pk,
 	gnutls_datum_t der = {NULL, 0};
 	ck_key_type_t key_type;
 
+	PKCS11_CHECK_INIT;
+
 	memset(&sinfo, 0, sizeof(sinfo));
 
 	ret = pkcs11_url_to_info(url, &info);
@@ -815,6 +856,7 @@ gnutls_pkcs11_privkey_generate2(const char *url, gnutls_pk_algorithm_t pk,
 		gnutls_pkcs11_obj_deinit(obj);
 	if (pkey != NULL)
 		gnutls_pubkey_deinit(pkey);
+
 	if (sinfo.pks != 0)
 		pkcs11_close_session(&sinfo);
 	gnutls_free(der.data);
