@@ -326,7 +326,7 @@ int _gnutls_find_config_path(char *path, size_t max_size)
 	}
 
 #ifdef _WIN32
-	while(0) {
+	if (home_dir == NULL || home_dir[0] == '\0') {
 		const char *home_drive = getenv("HOMEDRIVE");
 		const char *home_path = getenv("HOMEPATH");
 
@@ -337,20 +337,21 @@ int _gnutls_find_config_path(char *path, size_t max_size)
 		}
 	}
 #elif defined(HAVE_GETPWUID_R)
-	while(0) {
+	if (home_dir == NULL || home_dir[0] == '\0') {
 		struct passwd *pwd;
 		struct passwd _pwd;
+		int ret;
 		char tmp[512];
 
-		getpwuid_r(getuid(), &_pwd, tmp, sizeof(tmp), &pwd);
-		if (pwd != NULL) {
+		ret = getpwuid_r(getuid(), &_pwd, tmp, sizeof(tmp), &pwd);
+		if (ret == 0 && pwd != NULL) {
 			snprintf(path, max_size, "%s/" CONFIG_PATH, pwd->pw_dir);
 		} else {
 			path[0] = 0;
 		}
 	}
 #else
-	while(0) {
+	if (home_dir == NULL || home_dir[0] == '\0') {
 			path[0] = 0;
 	}
 #endif
@@ -467,42 +468,13 @@ int add_system_trust(gnutls_x509_trust_list_t list, unsigned int tl_flags,
 
 	return r;
 }
-#elif defined(ANDROID) || defined(__ANDROID__)
-#include <dirent.h>
-#include <unistd.h>
-static int load_dir_certs(const char *dirname,
-			  gnutls_x509_trust_list_t list,
-			  unsigned int tl_flags, unsigned int tl_vflags,
-			  unsigned type)
-{
-	DIR *dirp;
-	struct dirent *d;
-	int ret;
-	int r = 0;
-	char path[GNUTLS_PATH_MAX];
+#elif defined(ANDROID) || defined(__ANDROID__) || defined(DEFAULT_TRUST_STORE_DIR)
 
-	dirp = opendir(dirname);
-	if (dirp != NULL) {
-		do {
-			d = readdir(dirp);
-			if (d != NULL && d->d_type == DT_REG) {
-				snprintf(path, sizeof(path), "%s/%s",
-					 dirname, d->d_name);
+# include <dirent.h>
+# include <unistd.h>
 
-				ret =
-				    gnutls_x509_trust_list_add_trust_file
-				    (list, path, NULL, type, tl_flags,
-				     tl_vflags);
-				if (ret >= 0)
-					r += ret;
-			}
-		}
-		while (d != NULL);
-		closedir(dirp);
-	}
-
-	return r;
-}
+# if defined(ANDROID) || defined(__ANDROID__)
+#  define DEFAULT_TRUST_STORE_DIR "/system/etc/security/cacerts/"
 
 static int load_revoked_certs(gnutls_x509_trust_list_t list, unsigned type)
 {
@@ -534,6 +506,8 @@ static int load_revoked_certs(gnutls_x509_trust_list_t list, unsigned type)
 
 	return r;
 }
+# endif
+
 
 /* This works on android 4.x 
  */
@@ -543,21 +517,21 @@ int add_system_trust(gnutls_x509_trust_list_t list, unsigned int tl_flags,
 {
 	int r = 0, ret;
 
-	ret =
-	    load_dir_certs("/system/etc/security/cacerts/", list, tl_flags,
-			   tl_vflags, GNUTLS_X509_FMT_PEM);
+	ret = gnutls_x509_trust_list_add_trust_dir(list, DEFAULT_TRUST_STORE_DIR,
+		NULL, GNUTLS_X509_FMT_PEM, tl_flags, tl_vflags);
 	if (ret >= 0)
 		r += ret;
 
+# if defined(ANDROID) || defined(__ANDROID__)
 	ret = load_revoked_certs(list, GNUTLS_X509_FMT_DER);
 	if (ret >= 0)
 		r -= ret;
 
-	ret =
-	    load_dir_certs("/data/misc/keychain/cacerts-added/", list,
-			   tl_flags, tl_vflags, GNUTLS_X509_FMT_DER);
+	ret = gnutls_x509_trust_list_add_trust_dir(list, "/data/misc/keychain/cacerts-added/",
+		NULL, GNUTLS_X509_FMT_DER, tl_flags, tl_vflags);
 	if (ret >= 0)
 		r += ret;
+# endif
 
 	return r;
 }
